import torch
import numpy as np
from enum import Enum


class Norms(Enum):
    POPULATION = 0
    MIN_MAX = 1
    CONSTANT = 2


class PandemicDataset:
    def __init__(self,
                 name: str,
                 group_names: list,
                 N: int,
                 t,
                 *groups,
                 norm_name=Norms.MIN_MAX,
                 C=10**5,
                 use_scaled_time=False):
        """Class to hold all data for one training process.

        Args:
            group_names (list): Names of the groups, in which the population is split.
            N (int): Size of the population.
            t (np.array): Array of timesteps.
            *groups (np.array): Arrays of size data for each group for each timestep.
        """
        if torch.cuda.is_available():
            self.device_name = 'cuda'
        else:
            self.device_name = 'cpu'

        match norm_name:
            case Norms.POPULATION:
                self.__norm = self.__population_norm
                self.__denorm = self.__population_denorm
            case Norms.MIN_MAX:
                self.__norm = self.__min_max_norm
                self.__denorm = self.__min_max_denorm
            case Norms.CONSTANT:
                self.__norm = self.__constant_norm
                self.__denorm = self.__constant_denorm
            case _:
                self.__norm = self.__min_max_norm
                self.__denorm = self.__min_max_denorm

        self.name = name
        self.N = N
        self.t_init = t.min()
        self.t_final = t.max()
        self.C = C

        self.t_raw = torch.tensor(t, requires_grad=True, device=self.device_name)

        self.t_scaled = ((self.t_raw - self.t_init) / (self.t_final - self.t_init)).detach().requires_grad_()
        self.use_scaled_time = use_scaled_time
        if use_scaled_time:
            self.t_batch = self.t_scaled.view(-1, 1).float()
        else:
            self.t_batch = self.t_raw.view(-1, 1).float()

        self.__group_dict = {}
        for i, name in enumerate(group_names):
            self.__group_dict.update({name: i})

        self.__group_names = group_names

        self.__groups = [torch.tensor(group, device=self.device_name) for group in groups]

        self.__mins = [torch.min(group) for group in self.__groups]
        self.__maxs = [torch.max(group) for group in self.__groups]
        self.__norms = self.__norm(self.__groups)

    @property
    def number_groups(self):
        return len(self.__group_names)

    @property
    def data(self):
        return self.__groups

    @property
    def group_names(self) -> np.ndarray:
        return self.__group_names

    def __population_norm(self, data):
        return [(data[i] / self.N) for i in range(self.number_groups)]

    def __population_denorm(self, data):
        return [(data[i] * self.N) for i in range(self.number_groups)]

    def __min_max_norm(self, data):
        return [(data[i] - self.__mins[i]) / (self.__maxs[i] - self.__mins[i]) for i in range(self.number_groups)]

    def __min_max_denorm(self, data):
        return [(self.__mins[i] + (self.__maxs[i] - self.__mins[i]) * data[i]) for i in range(self.number_groups)]

    def __constant_norm(self, data):
        return [(data[i] / self.C) for i in range(self.number_groups)]

    def __constant_denorm(self, data):
        return [(data[i] * self.C) for i in range(self.number_groups)]

    def get_normalized_data(self, data: list):
        assert len(data) == self.number_groups, f'data parameter needs same length as there are groups in the dataset ({self.number_groups})'
        return self.__norm(data)

    def get_denormalized_data(self, data: list):
        assert len(data) == self.number_groups, f'data parameter needs same length as there are groups in the dataset ({self.number_groups})'
        return self.__denorm(data)

    def get_group(self, name: str):
        return self.__groups[self.__group_dict[name]]

    def get_min(self, name: str):
        return self.__mins[self.__group_dict[name]]

    def get_max(self, name: str):
        return self.__maxs[self.__group_dict[name]]

    def get_norm(self, name: str):
        return self.__norms[self.__group_dict[name]]
